@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2008-2020. All Rights Reserved.
+%% Copyright Ericsson AB 2008-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -51,22 +51,6 @@
 
 -define(NULL, 0).
 
--export_type([cert_opt/0, chain_opts/0, conf_opt/0,
-              test_config/0, test_root_cert/0]).
-
--type cert_opt()  :: {digest, public_key:digest_type()} | 
-                     {key, public_key:key_params() | public_key:private_key()} | 
-                     {validity, {From::erlang:timestamp(), To::erlang:timestamp()}} |
-                     {extensions, [#'Extension'{}]}.
--type chain_end()   :: root | peer.
--type chain_opts()  :: #{chain_end() := [cert_opt()],  intermediates =>  [[cert_opt()]]}.
--type conf_opt()    :: {cert, public_key:der_encoded()} | 
-                       {key,  public_key:private_key()} |
-                       {cacerts, [public_key:der_encoded()]}.
--type test_config() ::
-        #{server_config := [conf_opt()],  client_config :=  [conf_opt()]}.
--type test_root_cert() ::
-        #{cert := binary(), key := public_key:private_key()}.
 %%====================================================================
 %% Internal application APIs
 %%====================================================================
@@ -260,12 +244,12 @@ validate_extensions(OtpCert, ValidationState, UserState, VerifyFun) ->
 	    {ValidationState, UserState}
     end.
 %%--------------------------------------------------------------------
--spec normalize_general_name({rdnSequence, term()}) -> {rdnSequence, term()}. 
+-spec normalize_general_name({rdnSequence, term()}| binary()) -> {rdnSequence, term()}. 
 %%
 %% Description: Normalizes a general name so that it can be easily
 %%              compared to another genral name. 
 %%--------------------------------------------------------------------	
-normalize_general_name({rdnSequence, Issuer}) ->
+normalize_general_name({rdnSequence, Issuer}) ->    
     NormIssuer = do_normalize_general_name(Issuer),
     {rdnSequence, NormIssuer}.
 
@@ -468,11 +452,11 @@ match_name(Fun, Name, PermittedName, [Head | Tail]) ->
     end.
 
 %%%
--spec gen_test_certs(#{server_chain:= chain_opts(),
-                       client_chain:= chain_opts()} |
-                     chain_opts()) ->
-                            test_config() |
-                            [conf_opt()].
+-spec gen_test_certs(#{server_chain:= public_key:chain_opts(),
+                       client_chain:= public_key:chain_opts()} |
+                     public_key:chain_opts()) ->
+                            public_key:test_config() |
+                            [public_key:conf_opt()].
 %%
 %% Generates server and and client configuration for testing
 %% purposes. All certificate options have default values
@@ -537,17 +521,17 @@ gen_test_certs(
 
 
 x509_pkix_sign_types(#'SignatureAlgorithm'{algorithm = ?'id-RSASSA-PSS',
-                                           parameters = #'RSASSA-PSS-params'{hashAlgorithm = #'HashAlgorithm'{algorithm = Alg}}}) ->
+                                           parameters = #'RSASSA-PSS-params'{saltLength = SaltLen, hashAlgorithm = #'HashAlgorithm'{algorithm = Alg}}}) ->
     Hash = public_key:pkix_hash_type(Alg),
     {Hash, rsa_pss_pss, [{rsa_padding, rsa_pkcs1_pss_padding},
-                         {rsa_pss_saltlen, -1},
+                         {rsa_pss_saltlen, SaltLen},
                          {rsa_mgf1_md, Hash}]};
 x509_pkix_sign_types(#'SignatureAlgorithm'{algorithm = Alg}) ->
     {Hash, Sign} = public_key:pkix_sign_types(Alg),
     {Hash, Sign, []}.
 
 %%%
--spec root_cert(string(), [cert_opt()]) -> test_root_cert().
+-spec root_cert(string(), [public_key:cert_opt()]) -> public_key:test_root_cert().
 %%
 %% Generate a self-signed root cert
 root_cert(Name, Opts) ->
